// FILE: CLEANDIR.CPP
// MODULE: MAIN
// PROGRAM: CLEARDIR
// Compile using SMALL model

///////////////////////////////////////////////////////////////////////////////
//
//   CLEANDIR is a filter that reads input from STDIN, removes all lines that
//   begin with spaces or that contain the "<" character, and write the passed
//   lines to STDOUT.  This removes everything from a directory listing except
//   the listing lines.  It has a switch "/H" or "/h" to invoke a help page, and
//   another, "/K" or "/k" that keeps the line that identifies the directory.
//   That line begins with a space, which "/K" replaces with a "! !" so that it
//   will sort before the listing lines in most sort schemes.  "/k" just passes it.
//   CLEANDIR does not strip blank lines.
//
//   The program also performs the somewhat unrelated function of line truncation:
//   if it is given a "/X" or "/x" switch with a number, it will truncate its output
//   with that number column.  "/X" will cause it to accept ALL lines.
//
//   Maximum line length is 132 characters.
//
///////////////////////////////////////////////////////////////////////////////

#include <fstream.h>
#include <stdio.h>
#include <string.h>

#define max_len 133

void main(int argc, char * argv[])
{
int i, flag = 0, keep_flag = 0, bangbang = 0, filter_flag = -1, removebang = 0;
int short_flag = 0;
int line_len = max_len - 1;
char string[max_len + 1];
char * str;
    // parse the command tail
    for(i = 1; i < argc; i++)
    {
        if((argv[i][0] == '/') || (argv[i][0] == '-'))
        {
            switch(argv[i][1])
            {
                case 'H':       // HELP
                case 'h':
                    flag = -1;
                    cout << "\n CLEANDIR - FREEWARE Copyright 1991, Ted Davis\n";
                    cout << "\n A filter to remove all but the listing lines from the output of the\n";
                    cout << " DOS DIR command.  It accepts two switches, the /H or /h that got you this\n";
                    cout << " and /K or /k which causes it to pass the line that identifies the directory.\n";
                    cout << " /K adds ! ! to the start of the line so that it will often sort as the first line\n";
                    cout << " /k just passes the line.  This program is intended for use in pipes, as:\n";
                    cout << " dir | cleandir /K | csrts /$2.3 /$1 > dirfile <Enter>\n";
                    cout << " The program also truncates lines. /X{number} causes it to accept all lines up to\n";
                    cout << line_len << " characters and cut them off at the given column.  /x does the same except that\n";
                    cout << " it does not override the normal filters or truncate the directory name line.\n";
                    cout << " /R or /r will remove the ! ! marker from /K and is compatable with /X, but none of the\n";
                    cout << " other switches.  /S or /s will shorten the ID line to just the path.\n";
                    cout << " T.E.D. 4/29/91\n";
                    break;
                case 'K':       // keep ID line
                    bangbang = -1;  // mark it with ! !
                case 'k':
                    keep_flag = -1;
                    break;
                case 'R':           // remove ! ! marker
                case 'r':
                    removebang = filter_flag = -1;
                    break;
                case 'S':           // trim off "Directory of" from the ID line.
                case 's':
                    short_flag = -1;
                    break;
                case 'X':       // truncate line to given number, defeat filters.
                    filter_flag = 0;
                case 'x':       // truncate
                    str = argv[i] + 2;
                    sscanf(str, "%i", &line_len);
                    break;
                default:
                    cerr << "\n CLEARDIR - FREEWARE Copyright 1991, Ted Davis\n";
                    cerr << "\n ERROR: UNRECOGNIZED COMMAND: " << argv[i] << "\n";
                    flag = -1;
                    break;
            }
        }
    }
    // main business
    if(!flag)
    {
        while(!cin.eof())
        {
            cin.getline(string, max_len);
            if(filter_flag)
            {
                if(removebang && (string[0] == '!'))
                {
                    str = string + 3;
                    strcpy(string, str);
                    keep_flag = -1;
                }
                if((string[2] == 'i') && keep_flag)  // the only lower case 'i' in the third position is
                {                       // in the Directory of ... line.
                    if(short_flag)
                    {
                        str = string + 14;   // remove "Directory of"
                        strcpy(string, str);
                    }
                    if(bangbang)
                    {
                        cout << "! !" << string << "\n";
                        string[2] = '\0';            // prevents ! ! on next line due to leftover characters.
                    }
                    else
                    {
                        cout << string << "\n";
                    }
                }
                else
                {
                    if((string[0] != ' ') && (string[13] != '<'))
                    {
                        string[line_len] = '\0';     // truncate line
                        cout << string << "\n";
                        string[0] = '\0';      // prevents problems if next line does not write over these chars.
                        string[13] = '\0';
                    }
                }
            }
            else
            {
                string[line_len] = '\0';     // truncate line
                cout << string << "\n";
            }
        }
    }
}