// FILE: CSRTS1.CPP
// MODULE: MAIN
// PROGRAM: CSRTS
// Copyright 1991, Ted Davis, except a portion Copyright 1991, by McGraw-Hill
// Compile using compact model, there are two VERY large arrays.
// * Note that compact is the smallest model that allows use of huge (a form of far) 
// * strings with the string functions; models that use near data expect the strings
// * to be near also.

/////////////////////////////////////////////////////////////////////////////
//
//  Function:
//      Main - contains the command tail parser.  Initializes global variables
//
//
// *  In the interest of speed, this program uses C type arrays instead of the
// *  C++ Array class - huge pointers are slow enough, without having to go
// *  through function calls (possibly nested) to get at the individual
// *  characters.  Also with speed in mind, the string array is not sorted,
// *  an array of HUGE POINTERS into the array is sorted.  I am not even sure
// *  that the Array class can create huge arrays.
// *  In the interest of maximum file size, the arrays are huge; the source
// *  strings and sort patterns are packed into the array and each string
// *  is assigned an individual pointer to its first character.  This is similar
// *  to the way BASIC stores strings, except that the strings are null terminated
// *  and the descriptor consists of the pointer only, without the length.
// *
/////////////////////////////////////////////////////////////////////////////
//
// * Huge pointers occur, and are discussed, in the files CSRTS1.CPP, CSRTS3.CPP,
// * and CSRTS8.CPP.  The huge pointer comments are marked // *.

#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <conio.h>
#include <dos.h>
#include <alloc.h>


#define ShrinkFactor 1.3
// * the following defines express assumptions about the number of characters
// * per line and per sort pattern that are almost certainly wrong; but some
// * assumptions must be made for the memory to be divided between the arrays.
// * they can be overridden by the user.
#define default_chr_line 60
#define default_chr_sort 25
#define max_string_len  134


char c_notice[] = " Copyright 1991, Ted Davis\
                   Sort routine after Comb Sort,\
                   Copyright 1991, McGraw-Hill";
char string[256];               // a general purpose string variable.
char year[6];                   // stores the DOS year value as a string.
char * str;                     // a pointer used mostly to shorten strings.
int start_col = 1;              // an index to the first column to be sorted.
int field_width = max_string_len;   // the max number of characters to be sorted.
int flag = 0;                   // the go-no-go flag.
int i, k, l, j = 0;             // general use integer variables
int mi = 0, si = 0;             // indexes into the main and sort arrays.
int case_flag = 0;              // the "ignore case" flag.
int reverse_flag = 0;           // the master reverse sort flag.
int justify_flag = 0;           // right justification flag.
int left_flag = 0;              // left justification flag.
int blank_flag = 0;             // flag to force acceptance of blank lines.
int time_i = -1;                // index into the list of fields that are times.
int date_i = -1;                // index into the list of fields that are dates.
int Euro_flag = 0;              // marks dates as European (dd mm) sequence.
int reverse_i = -1;             // index into the list of fields that are to be reverse sorted.
int case_i = -1;                // index into the list of fields that are to be case insensitive.
int numbers_i = -1;             // index into the list of fields that are to be sorted as numbers.
int field_i = -1;               // index into the list of fields to be sorted.
int field_0 = 0;                // flags that $0 was used to signal use of entire line.
char * p;                       // another string pointer line str.
int numbers[3] = {-1, -1, -1};  // temporary storage of command tail numbers.
int field_numbers[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of fields to sort.
int fields_widths[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of field widths.
int fields_decimal[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of decimal field widths.
int time_fields[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of fields that are times.
int date_fields[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of fields that are dates.
int reverse_fields[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of fields that are to be reverse sorted.
int case_fields[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of fields that are to be case insensitive.
int number_fields[8] = {-1,-1,-1,-1,-1,-1,-1,-1};
                                // the list of fields that are to be numbers.
char fields[8][82];             // the fields to be sorted are stored here as strings.
// * the following huge pointers are four byte blocks in the normal data area
// * and are initialized to NULL by the compiler.
char huge *main_array;          // * this one will mark the beginning of the
                                // * memory block used for string storage.
char huge *strn;                // * a general purpose pointer used as an interface
                                // * between the strings in the array and the C string
                                // * functions (which don't like expressions as arguments).
                                // * strn is used for the strings read from the source.
char huge *sort_str;            // * this is the same as strn except that it is used for
                                // * the pattern string upon which the sort is based.
                                // * its value is used as the start of the much smaller index array.
union                           // * this union allows declaring the index array as a huge pointer
{                               // * to type char, then using the space as an array of huge pointers.
    void far * prime_array;     // * this pointer is used only to allocate all the free memory;
    char huge *ix;              // * the same pointer as if to a char array.
    char * huge *index_array;   // * the same pointer as if to an array of huge pointers that occupies
                                // * the same space.
} ptr;                          // * the name of the union.
// * since the arrays may need indexes larger than can be accommodated by type int, the variables
// * that are used to compute indexes and pointers are made type long.
long array_size, long_i = 0, long_x = 0, long_j = 0;
                                // * array_size is used to keep track of the upper limit of memory.
                                // * long_i is used as the index into the index array; long_x and
                                // * long_j are used with the string array.
// * the next two longs are used to divide up the available memory into the string array and the index array.
long chr_line = default_chr_line;   // * estimated average characters per source line.
long chr_sort = default_chr_sort;   // * estimated average characters per sort pattern.
                                    // * those two values can be overridden by the user from the command line.
long max_index;                 // * sets the upper limit on how many lines can be sorted.  It is
                                // * the maximum number of entries in the index array.  the value is to be
                                // * calculated when the arrays are created.



extern void DoubleCombSort(long size, int direction,  char * huge * index);
extern void time_convert(char *string, char *extra);
extern void date_convert(char *string, int flag, char *year);
extern void reverse_convert(char *string);
extern void print_copyright(void);
extern void print_copyright_err(void);
extern void print_help(void);
extern int get_field(char * args, int fields[], int flag);
extern int read_field( int * fields, int _i);
extern void crunch_numbers(void);
extern void load_array(void);


void main(int argc, char *argv[])
// Contains most of the overhead code for the program.  It parses the command tail
// for valid commands, generates error and warning messages if the commands are
// defective or conflicting.  It calls the functions that read the file, process
// user data and massage the fields, perform the sort, and output the results.
{
struct date now;                // used to get the year for date conversions.

    // parse the command tail
    for(i = 1; i < argc; i++)
    {
        k = 0;
        if((argv[i][0] == '/') || (argv[i][0] == '-'))
        {
            switch(argv[i][1])
            {
                case '+':       // column number, up to two allowed
                    if(numbers[0] == -1)
                    {
                        str = argv[i] + 2;
                        sscanf(str, "%i", &k);  // get the number
                        numbers[0] = abs(k);    // ignore minus signs
                        if(k == 0)
                        {
                            print_copyright_err();
                            cerr << "\nFATAL ERROR: DEFECTIVE COLUMN ARGUMENT - " << argv[i] << "\n";
                            flag = -1;
                        }
                    }
                    else
                    {
                        if(numbers[1] == -1)
                        {
                            str = argv[i] + 2;
                            sscanf(str, "%i", &k);
                            numbers[1] = abs(k);
                            if(k == 0)
                            {
                                print_copyright_err();
                                cerr << "\nFATAL ERROR: DEFECTIVE COLUMN ARGUMENT - " << argv[i] << "\n";
                                flag = -1;
                            }
                        }
                        else
                        {
                            print_copyright_err();
                            cerr << "\nFATAL ERROR: TOO MANY COLUMN ARGUMENTS - " << argv[i] << "\n";
                            flag = -1;
                        }
                    }
                    if((field_i != -1) && (numbers[1] != -1))
                    {
                        numbers[1] = -1;
                        print_copyright_err();
                        cerr << "\nWARNING: SECOND COLUMN ARGUMENT CONFLICTS WITH FIELD NUMBER, "<< argv[i] <<" IGNORED\n";
                    }
                    if((numbers[2] != -1) && (numbers[1] != -1))
                    {
                        numbers[0] = numbers[1] = -1;
                        print_copyright_err();
                        cerr << "\nWARNING: SECOND COLUMN ARGUMENT CONFLICTS WITH FIELD WIDTH, "<< argv[i] <<" IGNORED\n";
                    }
                    break;
                case '$':           // field number, up to 8 allowed, used in order of occurrence
                    field_i = get_field(argv[i], field_numbers, field_i);
                    break;
                case 'B':               // accept empty lines
                case 'b':
                    blank_flag = -1;
                    break;
                case 'C':               // flags for case sensitivity
                case 'c':
                    case_flag = -1;  // case sort everything or flag for individual
                    if(strlen(argv[i]) < 2)
                    {
                        str = argv[i] + 2;
                        case_i = read_field(case_fields, case_i);
                    }
                    break;
                case 'D':                   // given field is a date
                case 'd':                   // negative numbers mean European format
                // get the year
                    getdate(&now);
                // convert year to a string
                    itoa(now.da_year, year, 10);
                // get field number
                    field_i = get_field(argv[i], field_numbers, field_i);
                // mark as date field
                    date_fields[++date_i] = abs(k) - 1;
                    if(k < 0) { Euro_flag = -1;}
                    break;
                case 'J':               // right justify in field
                case 'j':
                    justify_flag = -1;
                    break;
                case 'H':               // HELP
                case 'h':
                    print_help();
                    flag = -1;
                    break;
                case 'L':               // left justify line
                case 'l':
                    left_flag = -1;
                    break;
                case 'N':               // numbers fields
                case 'n':
                    field_i = get_field(argv[i], field_numbers, field_i);
                    number_fields[++numbers_i] = abs(k) - 1;
                    break;
                case 'O':               // override default assumptions about
                case 'o':               // char/line and char/pattern.
                    p = strchr(argv[i], '.');
                    if(p)
                    {
                        j = p - argv[i];
                        argv[i][j] = ' ';
                        sscanf(argv[i], "%l%l", &chr_line, &chr_sort);
                        // THERE IS NO ERROR CHECKING FOR THIS COMMAND
                    }
                    break;
                case 'R':               // reverse sort
                case 'r':
                    if(strlen(argv[i]) == 2)
                    {
                        reverse_flag = -1;  // reverse sort everything
                    }
                    else    // there is a numerical argument, reverse sort given field
                    {
                        str = argv[i] + 2;
                        sscanf(str, "%i", &k);
                        if(k)
                        {
                            reverse_fields[++reverse_i] = abs(k) - 1;
                        }
                        else
                        {
                            print_copyright_err();
                            cerr << "\nFATAL ERROR: DEFECTIVE REVERSE SORT FIELD ARGUMENT - " << argv[i] << "\n";
                        }
                    }
                    break;
                case 'T':               // flag field as times
                case 't':
                    field_i = get_field(argv[i], field_numbers, field_i);
                    time_fields[++time_i] = abs(k) - 1;
                    break;
                case '0':               // field width, one allowed
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    if(numbers[2] != -1)
                    {
                        print_copyright_err();
                        cerr <<"\nFATAL ERROR: TOO MANY FIELD WIDTH ARGUMENTS - "<< argv[i] <<"\n";
                        flag = -1;
                    }
                    else
                    {
                        str = argv[i] + 1;
                        sscanf(str, "%i", &k);
                        numbers[2] = abs(k);
                        if(k == 0)
                        {
                            print_copyright_err();
                            cerr << "\nFATAL ERROR: DEFECTIVE FIELD WIDTH ARGUMENT - " << argv[i] << "\n";
                            flag = -1;
                        }
                    }
                    if(field_i != -1)
                    {
                        numbers[2] = -1;
                        print_copyright_err();
                        cerr << "\nWARNING: FIELD WIDTH CONFLICTS WITH FIELD NUMBER, "<< argv[i] <<" IGNORED\n";
                    }
                    if(numbers[1] != -1)
                    {
                        numbers[2] = -1;
                        print_copyright_err();
                        cerr << "\nWARNING: FIELD WIDTH CONFLICTS WITH TWO COLUMN ARGUMENTS, "<< argv[i] <<" IGNORED\n";
                    }
                    break;
                default:                // ERRORS in command tail
                    print_copyright_err();
                    cerr <<"\nFATAL ERROR: UNRECOGNIZED ARGUMENT "<< argv[i] <<"\n";
                    flag = -1;
                    break;
            }
        }
        else                            // neither / nor - present
        {
            print_copyright_err();
            cerr << "\nFATAL ERROR: DEFECTIVE ARGUMENT " << argv[i] << "\n";
            flag = -1;
        }
        if(flag) {break;}
    }
    crunch_numbers();

    // all necessary numbers are massaged, on to reading the file via redirected STDIN

    load_array();
    if(!flag)               // don't sort if there are errors
    {
        // the main business of the program is here
        DoubleCombSort(long_i, reverse_flag, ptr.index_array);
        for(long_j = 0; long_j < long_i; long_j+=2) // send the sorted array to STDOUT (redirected),
        {                       // one line at a time and add a newline marker
            strn = ptr.index_array[long_j];
            cout << strn << '\n';
        }
    }

    // release the memory allocated with farmalloc in load_array.
    farfree(ptr.prime_array);

}
